import uuid
from threading import Thread
from typing import (
    Callable,
    ClassVar,
    Dict,
    Generic,
    Optional,
    Tuple,
    TypeVar,
)

from pandos.futures.enums import FutureStatus
from pandos.futures.results import FutureResult


A = TypeVar("A")


class FutureStore(Generic[A]):
    _instance: ClassVar[Optional['FutureStore']] = None
    registry: ClassVar[Dict[str, FutureResult]] = {}
    # Undefined instance attributes
    expiration_once_realized_in_seconds: Optional[int]
    max_size: Optional[int]
    max_age: int

    # Singleton Pattern
    def __new__(
            cls,
            expiration_once_realized_in_seconds: Optional[int] = None,
            max_size: Optional[int] = None,
            max_age: int = 100_000,
    ):
        if cls._instance is None:
            cls._instance = super(FutureStore, cls).__new__(cls)
            # Instance attributes here:
            cls._instance.expiration_once_realized_in_seconds = expiration_once_realized_in_seconds
            cls._instance.max_size = max_size
            cls._instance.max_age = max_age
        return cls._instance

    @property
    def successes(self) -> int:
        return len([future for future in self.registry.values() if future.status == FutureStatus.SUCCESS])

    @property
    def failures(self) -> int:
        return len([future for future in self.registry.values() if future.status == FutureStatus.FAILURE])

    @property
    def unrealized(self) -> int:
        return len([future for future in self.registry.values() if future.status == FutureStatus.PENDING])

    def register(self, function: Callable[..., A], *args, **kwargs) -> Tuple[Thread, str]:
        if self.max_size is not None and len(self.registry) > self.max_size:
            # Think about this... do we really need to drop old futures?
            [
                self.registry.pop(key)
                for key, val in self.registry.items()
                if val.age and val.age > self.max_age
            ]
        # Register future
        future_id = str(uuid.uuid4())
        self.registry[future_id] = FutureResult(status=FutureStatus.PENDING)
        # Create thread
        thread = Thread(target=lambda: self.registry[future_id].define_with(function, *args, **kwargs), daemon=True)
        thread.start()
        return thread, future_id



from functools import wraps
from dataclasses import dataclass
from typing import (
    Callable,
    Generic,
    Optional,
    TypeVar,
    Union,
)

from pandos.exceptions import pandos_exceptions
from pandos.futures.store import FutureStore
from pandos.futures.enums import FutureStatus
from pandos.monads.either import (
    Either,
    Right,
    Left,
)


A = TypeVar("A")
B = TypeVar("B")


@dataclass(frozen=True, slots=True)
class FutureFunctionWrapper:
    target: Callable[..., A]

    def future(self, *args, **kwargs) -> 'Future[A]':
        return Future[A](self.target, *args, **kwargs)

    def __call__(self, *args, **kwargs) -> A:  # type: ignore
        return self.target(*args, **kwargs)


class Future(Generic[A]):
    __slots__ = "thread", "future_id"

    def __init__(self, function: Callable[..., A], *args, **kwargs):
        self.thread, self.future_id = FutureStore().register(function, *args, **kwargs)

    def __str__(self) -> str:
        return self.future_id

    def __repr__(self) -> str:
        match FutureStore.registry[self.future_id].status:
            case FutureStatus.SUCCESS:
                return "Future[Success]"
            case FutureStatus.FAILURE:
                return "Future[Failure]"
            case FutureStatus.PENDING:
                return "Future[Pending]"
        # For some reason, the type check is expecting a return... raising an error instead
        return pandos_exceptions.FUTURES_INVALID_STATE_ERROR.throw()

    def __iter__(self):
        yield from [self.wait_and_resolve()]

    @classmethod
    def comprehension(cls, generator, timeout: Optional[int] = None):
        return Future(function=lambda: next(generator), timeout=timeout)

    @classmethod
    def decorator(cls, timeout: Optional[int] = None) -> Callable[[Callable[..., A]], FutureFunctionWrapper]:
        return lambda function: wraps(function, updated=())(FutureFunctionWrapper)(target=function)  # type: ignore

    @property
    def status(self) -> FutureStatus:
        return FutureStore.registry[self.future_id].status

    def _coerce_current_value(self) -> Either[A]:
        result = FutureStore.registry[self.future_id]
        match result.status:
            case FutureStatus.SUCCESS:
                return result.value
            case FutureStatus.FAILURE:
                print("Failure")
                return result.value
            case FutureStatus.PENDING:
                return pandos_exceptions.FUTURES_UNRESOLVED_ERROR.throw(
                    f"Trying to coerce value from unresolved future: {self.future_id}"
                )

    def wait(self, timeout: Optional[float] = None) -> Either[A]:
        self.thread.join(timeout=timeout)
        try:
            return self._coerce_current_value()
        except pandos_exceptions.FUTURES_UNRESOLVED_ERROR.value:
            return pandos_exceptions.FUTURES_INCONSISTENT_ERROR.throw(
                f"Future with inconsistent state detected in wait-call (should be resolved): {self.future_id}"
            )

    def wait_and_resolve(self, timeout: Optional[float] = None) -> A:
        return self.wait(timeout=timeout).resolve()

    def get(
            self,
            or_else: Optional[B] = None,
            blocking: bool = False,
            timeout:  Optional[float] = None
    ) -> Union[Either[A], Optional[B]]:
        # TODO: Should or_else apply only on unresolved (pending) futures or also on failed futures?
        if blocking:
            return self.wait(timeout=timeout)
        try:
            return self._coerce_current_value()
        except pandos_exceptions.FUTURES_UNRESOLVED_ERROR.value:
            # TODO: Logger?
            return or_else

    @classmethod
    def from_value(cls, x: A) -> 'Future[A]':
        return Future[A](lambda: x)

    def flat_map(self, function: Callable[[A], 'Future[B]'], timeout: Optional[float] = None) -> 'Future[B]':
        # This works, but is blocking: return function(self.wait_and_resolve())
        # The following implementation is just a placeholder and should be improved for performance.
        return Future(
            function=lambda: function(self.wait_and_resolve(timeout=timeout)).wait_and_resolve(timeout=timeout)
        )

    def map(self, function: Callable[[A], B], timeout: Optional[float] = None) -> 'Future[B]':
        return Future(
            function=lambda: function(self.wait_and_resolve(timeout=timeout))
        )



import os
import datetime as dt
from time import perf_counter
from dataclasses import dataclass
from types import TracebackType
from typing import (
    Any,
    Optional,
    Type,
)

from pandos.version import Version, version
from pandos.maturity import MaturityLevel
from pandos.settings import get_logger


logger = get_logger(name=__name__)


MaturityLevel.ALPHA.set_module(
    file=__file__,
    logger=logger
)


@dataclass(frozen=True, slots=True)
class CLI:
    version: Version = version
    _start_performance_counter: float = perf_counter()
    _now: dt.datetime = dt.datetime.utcnow()

    @property
    def duration(self) -> float:
        return perf_counter() - self._start_performance_counter

    @property
    def now(self) -> str:
        return self._now.isoformat()

    @property
    def about(self) -> str:
        return f"Pandos Version: {self.version.value}"

    def environ(self, varname: str) -> Optional[str]:
        return os.environ.get(varname)

    def feature_maturity(self, module: str):
        # TODO: Improve implementation...
        #       ATM we are just triggering the log function by importing the module by name
        import importlib

        importlib.import_module(module)

    def hello(self, name: Optional[str] = None):
        name = name or "world"
        return f"Hello, {name}!"

    def __enter__(self):
        logger.debug("Entering Pandos CLI context manager")
        logger.info(self.about)
        logger.info("Pandos start timestamp %s", self.now)
        return self

    def __exit__(
            self,
            exc_type: Optional[Type[Exception]] = None,
            exc_value: Optional[Exception] = None,
            exc_tb: Optional[TracebackType] = None,
    ):
        duration = self.duration
        status = "success" if exc_value is None else "failure"
        logger.info("Pandos command execution status: %s", status)
        logger.info("Pandos command execution duration: %f", duration)

    def execute(self, command: str, *args, **kwargs) -> Any:
        with self as cli:
            executable = getattr(cli, command)
            logger.info("Pandos command: %s", command)
            out = executable(*args, **kwargs)
        return out

    @classmethod
    def system(cls, syscli: str) -> Type:
        import importlib

        module_name = ".".join(["pandos", "system", syscli, "syscli"])
        module_reference = importlib.import_module(module_name)
        cls_ext = getattr(module_reference, "CLIExtension")
        return type("CLI" + syscli.title(), (cls_ext, cls), {
            "pandos_version": cls().version.value,
        })



import enum
from typing import Any

from pandos.settings import get_logger
from pandos.maturity import MaturityLevel


logger = get_logger(name=__name__)


MaturityLevel.ALPHA.set_module(
    file=__file__,
    logger=logger,
)


# Inspired on:
# - https://github.com/python/cpython/blob/4d1f033986675b883b9ff14588ae6ff78fdde313/Lib/enum.py#L1265
# - https://docs.python.org/3.11/library/enum.html#enum.StrEnum
class TextEnum(str, enum.Enum):

    def __new__(cls, value: str):
        member = str.__new__(cls, value)
        member._value_ = value
        return member

    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list[Any]) -> Any:
        return name.lower()



import os
import enum
from logging import Logger
from dataclasses import dataclass
from typing import Optional

from pandos.settings import get_logger
from pandos.settings import (
    PANDOS_DISABLE_FEATURE_MATURITY_LOGS,
)


logger = get_logger(name=__name__)


# pandos-mypyc-ignore-file
# TODO: This file is ingored due to the following mypyc error:
#       Non-extension classes may not inherit from extension classes.
#       There's an on-going thread in github to fix this: https://github.com/python/mypy/issues/9612
@dataclass(frozen=True)
class FeatureMaturity:
    tag: str
    message: Optional[str]
    description: str
    loglevel: str

    @property
    def maturity_level(self) -> 'MaturityLevel':
        return MaturityLevel[self.tag]

    def set_module(
            self,
            file: str,
            logger: Logger
    ):
        # Early exit if the maturity level logs are disabled
        if PANDOS_DISABLE_FEATURE_MATURITY_LOGS:
            return
        # Get the module name from the reference file path
        root, part, module = file.partition(os.path.join("src", "main", "pandos", ""))
        if module.endswith("__init__.py"):
            module = os.path.dirname(module)
        # Verify if we have a message template
        if not self.message:
            if self.maturity_level.code < 1:
                from pandos.exceptions import pandos_exceptions  # Lazy import

                pandos_exceptions.FEATURE_MATURITY_LEVEL_UNDEFINED_MESSAGE_INCONSISTENCY.throw()
            return
        # Format the message template and pass it through the logs
        message = self.message.format(module=module)
        logger_function = getattr(logger, self.loglevel)
        logger_function(message)


class MaturityLevel(FeatureMaturity, enum.Enum):
    PRE_ALPHA = (
        "PRE_ALPHA",
        "You are using a pre-alpha feature: {module}",
        "Pre-Alpha status means that the feature still being figured out and in WIP status. "
        "This most likely results in incomplete/inconsistent features and non-stable interfaces. "
        "All changes regarding these features, even non-backward compatible changes, will be registered as patch.",
        "warning"
    )
    ALPHA = (
        "ALPHA",
        "You are using an alpha feature: {module}",
        "Alpha status means that the current feature is complete but still considered as experimental. "
        "Breaking changes can occur at any moment and documentation/testing is limited. "
        "All changes regarding these features, even non-backward compatible changes, will be registered as patch.",
        "warning"
    )
    BETA = (
        "BETA",
        "You are using a beta feature: {module}",
        "Beta status means that the feature is considered mature-enough to be used by technical & savvy users. "
        "At this point, the feature should be almost 100% complete but may still receive significant updates. "
        "We encourage users to submit feedback and bug reports since this will be the first time the feature will be "
        "used by a larger number of users. Technical documentation and unit-testing should be almost done by now. "
        "Consider that changes regarding these features, including non-backward compatible changes, "
        "will only affect the minor or patch values (`minor.patch`). If you see this message, you are still early "
        "to the development process! You can make a difference by letting us know your experience here: TBD",
        "warning"
    )
    GAMMA = (
        "GAMMA",
        None,
        "Gamma status means that the feature is considered mature-enough to be used by most users. "
        "The technical implementation of the feature is mostly over alongside with technical documentation "
        "and unit-testing. The feature is now being formally documented (end-user docs) and use-cases / examples are "
        "being made available. Extensive testing by our QA team is being done and semantic versioning is "
        "fully applied at this point (`major.minor.patch`).",
        "info"
    )
    STABLE = (
        "STABLE",
        None,
        "Stable status means that the feature is well-documented, has unit-testing, had passed QA and the "
        "interface/functionality are resilient to change. Changes are incorporated into the package-level versioning "
        "via semantic-versioning (`major.minor.patch`).",
        "debug"
    )
    TO_BE_DEPRECATED = (
        "TO_BE_DEPRECATED",
        "Danger zone! You are using a feature soon to be deprecated: {module}",
        "The 'To be Deprecated' status means that the feature (module) you are using will most likely be removed "
        "on the next major release.",
        "warning"
    )

    @property
    def code(self) -> int:
        match self:
            case self.PRE_ALPHA | self.ALPHA | self.BETA | self.TO_BE_DEPRECATED:
                return 0
            case self.GAMMA | self.STABLE:
                return 1
        from pandos.exceptions import pandos_exceptions  # Lazy import

        return pandos_exceptions.FEATURE_MATURITY_LEVEL_UNDEFINED.throw()


MaturityLevel.ALPHA.set_module(
    file=__file__,
    logger=logger
)



import os
import logging


PANDOS_DISABLE_FEATURE_MATURITY_LOGS = int(os.environ.get(
    "PANDOS_DISABLE_FEATURE_MATURITY_LOGS",
    default="0"
))


def get_logger(name: str) -> logging.Logger:
    logging.basicConfig(
        encoding="utc-8",
        level=logging.INFO,  # TODO: Parametrize log-level
    )
    return logging.getLogger(name)



from time import perf_counter
from dataclasses import dataclass
from typing import (
    Callable,
    Generic,
    Optional,
    TypeVar,
)
from pandos.futures.enums import FutureStatus
from pandos.monads.either import (
    Either,
    Right,
    Left,
)


A = TypeVar("A")


@dataclass(frozen=False, slots=True)
class FutureResult(Generic[A]):
    status: FutureStatus
    value: Optional[Either[A]] = None
    raise_exception: bool = False
    exec_duration: Optional[float] = None
    _ts_perf_counter: Optional[float] = None

    @property
    def age(self) -> Optional[float]:
        if self._ts_perf_counter is None:
            return None
        return perf_counter() - self._ts_perf_counter

    def define_with(self, function: Callable[..., A], *args, **kwargs):
        start = perf_counter()
        # The idea of the try-except statement is to wrap the result value on an Either Monad. Nonetheless, consider
        # that the function evaluation might return any of the following scenarios:
        # - Scenario 1 (ideal): Any python value or exception -- there should be managed with the Either Monad
        # - Scenario 2: An either monad instance... use that value directly and update the status accordingly.
        # - Scenario 3: Another future... Still thinking about how to handle this best.
        try:
            # Apply the function evaluation! Consider that this section should be running on a different thread
            # therefore it should be non-blocking.
            match function(*args, **kwargs):
                case right if isinstance(right, Right):
                    self.value = right
                    self.status = FutureStatus.SUCCESS
                case left if isinstance(left, Left):
                    self.value = left
                    self.status = FutureStatus.FAILURE
                case value:
                    self.value = Right.from_value(x=value)
                    self.status = FutureStatus.SUCCESS

        except Exception as e:
            self.value = Left(exception=e)
            self.status = FutureStatus.FAILURE
            if self.raise_exception:
                raise
        finally:
            self._ts_perf_counter = perf_counter()
            self.exec_duration = self._ts_perf_counter - start



import enum
from typing import ClassVar, Optional, Type

from pandos.settings import get_logger
from pandos.maturity import MaturityLevel


logger = get_logger(name=__name__)


MaturityLevel.ALPHA.set_module(
    file=__file__,
    logger=logger
)


class ExceptionEnumBuilder(enum.Enum):

    def throw(self, message: Optional[str] = None):
        raise self.value(message)

    @classmethod
    def members(cls, name: str, **kwargs):
        return cls(
            value=name,
            # type: ignore
            names=[
                (exc_name, exc_cls)
                for exc_name, exc_cls in kwargs.items()
            ]
        )


class PandosException(Exception):
    default_message: ClassVar[str] = "Pandos Generic Exception"

    def __init__(self, message: Optional[str] = None):
        super().__init__(message or self.default_message)

    @classmethod
    def custom(
            cls,
            exception_name: str,
            default_message: Optional[str] = None,
    ) -> Type:
        attrs = {
            "default_message": default_message or cls.default_message
        }
        return type(exception_name, (cls,), attrs)

    @classmethod
    def enums(cls):
        return ExceptionEnumBuilder.members(
            name="PandosExceptionCatalog",
            # Register exceptions here:
            PANDOS_BUILTIN_CUSTOM_EXCEPTION=cls.custom(
                exception_name="PandosBuiltInCustomException",
                default_message="This is an example custom exception - you should not find this error message in prod",
            ),
            MONADS_EITHER_INCONSISTENCY=cls.custom(
                exception_name="PandosMonadsEitherInconsistency",
                default_message="Inconsistency detected in Either Monad",
            ),
            FUTURES_UNRESOLVED_ERROR=cls.custom(
                exception_name="PandosFuturesUnresolvedError",
                default_message="The future has not been resolved yet."
            ),
            FUTURES_INCONSISTENT_ERROR=cls.custom(
                exception_name="PandosFuturesInconsistentError",
                default_message="An inconsistent future was detected... this should not happen."
            ),
            FUTURES_INVALID_STATE_ERROR=cls.custom(
                exception_name="PandosFuturesInvalidStateError",
                default_message="Future match-statement detected an invalid state... this should never happen.",
            ),
            FEATURE_MATURITY_LEVEL_UNDEFINED=cls.custom(
                exception_name="PandosFeatureMaturityLevelUndefined",
                default_message="Feature maturity level undefined... this should not happen!"
            ),
            FEATURE_MATURITY_LEVEL_UNDEFINED_MESSAGE_INCONSISTENCY=cls.custom(
                exception_name="PandosFeatureMaturityLevelUndefinedMessageInconsistency",
                default_message="The feature maturity level being used should have a message template."
            ),
        )


pandos_exceptions = PandosException.enums()
